
/******************************************************************************
* MODULE     : symbolic_arith.cpp
* DESCRIPTION: Arithmetic on symbolic expressions
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "symbolic_impl.hpp"
namespace mml {

symbolic
generic_add (const symbolic& s1, const symbolic& s2) {
  if (s1 == SYMB_ZERO) return s2;
  if (s2 == SYMB_ZERO) return s1;
  return symbolic_compound (SYMB_PLUS, s1, s2);
}

symbolic
generic_neg (const symbolic& s1) {
  return symbolic_compound (SYMB_MINUS, s1);
}

symbolic
generic_sub (const symbolic& s1, const symbolic& s2) {
  if (s1 == SYMB_ZERO) return -s2;
  if (s2 == SYMB_ZERO) return s1;
  return symbolic_compound (SYMB_MINUS, s1, s2);
}

symbolic
generic_mul (const symbolic& s1, const symbolic& s2) {
  if (s1 == SYMB_ZERO) return SYMB_ZERO;
  if (s2 == SYMB_ZERO) return SYMB_ZERO;
  if (s1 == SYMB_ONE) return s2;
  if (s2 == SYMB_ONE) return s1;
  return symbolic_compound (SYMB_TIMES, s1, s2);
}

symbolic
generic_div (const symbolic& s1, const symbolic& s2) {
  if (s1 == SYMB_ZERO) return SYMB_ZERO;
  if (s2 == SYMB_ONE) return s1;
  return symbolic_compound (SYMB_OVER, s1, s2);
}

symbolic
generic_pow (const symbolic& s1, const symbolic& s2) {
  if (s1 == SYMB_ZERO) return SYMB_ZERO;
  if (s1 == SYMB_ONE) return SYMB_ONE;
  if (s2 == SYMB_ZERO) return SYMB_ONE;
  if (s2 == SYMB_ONE) return s1;
  return symbolic_compound (SYMB_POWER, s1, s2);
}

symbolic
operator + (const symbolic& s1, const symbolic& s2) {
  switch (max (s1->get_add_type (), s2->get_add_type ())) {
  case ADD_INTEGER:
    return integer_add (s1, s2);
  case ADD_RATIONAL:
    return rational_add (s1, s2);
  case ADD_COMMUTATIVE:
    return sum_add (s1, s2);
  default:
    return generic_add (s1, s2);
  }
}

symbolic
operator - (const symbolic& s1, const symbolic& s2) {
  switch (max (s1->get_add_type (), s2->get_add_type ())) {
  case ADD_INTEGER:
    return integer_sub (s1, s2);
  case ADD_RATIONAL:
    return rational_sub (s1, s2);
  case ADD_COMMUTATIVE:
    return sum_sub (s1, s2);
  default:
    return generic_sub (s1, s2);
  }
}

symbolic
operator - (const symbolic& s1) {
  switch (s1->get_add_type ()) {
  case ADD_INTEGER:
    return integer_neg (s1);
  case ADD_RATIONAL:
    return rational_neg (s1);
  case ADD_COMMUTATIVE:
    return sum_neg (s1);
  default:
    return generic_neg (s1);
  }
}

symbolic
scalar_mul (const symbolic& sc, const symbolic& s1) {
  switch (max (sc->get_add_type (), s1->get_add_type ())) {
  case ADD_INTEGER:
    return integer_mul (sc, s1);
  case ADD_RATIONAL:
    return rational_mul (sc, s1);
  case ADD_COMMUTATIVE:
    return sum_scalar_mul (sc, s1);
  default:
    return generic_mul (sc, s1);
  }
}

symbolic
operator * (const symbolic& s1, const symbolic& s2) {
  switch (max (s1->get_mul_type (), s2->get_mul_type ())) {
  case MUL_INTEGER:
    return integer_mul (s1, s2);
  case MUL_RATIONAL:
    return rational_mul (s1, s2);
  case MUL_COMMUTATIVE:
    return product_mul (s1, s2);
  default:
    return generic_mul (s1, s2);
  }
}

symbolic
operator / (const symbolic& s1, const symbolic& s2) {
  switch (max (s1->get_mul_type (), s2->get_mul_type ())) {
  case MUL_INTEGER:
    return integer_div (s1, s2);
  case MUL_RATIONAL:
    return rational_div (s1, s2);
  case MUL_COMMUTATIVE:
    return product_div (s1, s2);
  default:
    return generic_div (s1, s2);
  }
}

symbolic
pow (const symbolic& s1, const symbolic& s2) {
  switch (max (s1->get_pow_type (), s2->get_pow_type ())) {
  case POW_EXPO_LINEAR:
    return product_pow (s1, s2);
  default:
    return generic_pow (s1, s2);
  }
}

} // namespace mml
