
/******************************************************************************
* MODULE     : symbolic_product.cpp
* DESCRIPTION: Commutative products of a finite number of symbolic expressions
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "symbolic_impl.hpp"
namespace mml {

class symbolic_product_rep: public symbolic_rep {
public:
  symbolic_commutative prod;

protected:
  // accessors and syntactic identity
  symbolic_type type () const { return SYMBOLIC_PRODUCT; }
  nat hash () const { return 741852963 ^ mml::hash (prod); }
  bool identical (symbolic_rep* rep) const {
    if (rep->type() != SYMBOLIC_PRODUCT) return false;
    return prod == ((symbolic_product_rep*) rep)->prod;
  }
  symbolic flatten (bool recursive) const {
    iterator<symbolic> it= entries (prod);
    symbolic acc= 0;
    while (busy (it)) {
      symbolic term;
      symbolic var  = *it;
      symbolic power= prod[var];
      if (power == 1) term= var;
      else term= symbolic_compound ("^", var, power);
      if (acc == 0) acc= term;
      else acc= symbolic_compound ("*", term, acc);
      ++it;
    }
    return mml::flatten (acc, recursive);
  }

  // arithmetic
  //add_type get_add_type () const { return ?; }
  mul_type get_mul_type () const { return MUL_COMMUTATIVE; }
  //pow_type get_pow_type () const { return ?; }

public:
  symbolic_product_rep (const symbolic_commutative& prod2): prod (prod2) {}
};

/******************************************************************************
* Constructors
******************************************************************************/

symbolic
symbolic_product (const symbolic_commutative& prod) {
  // cout << "product " << prod << "\n";
  switch (N(prod)) {
  case 0:
    return SYMB_ONE;
  case 1:
    {
      symbolic f= *entries (prod);
      if (prod[f] == SYMB_ONE) return f;
      return new symbolic_product_rep (prod);
    }
  default:
    return new symbolic_product_rep (prod);
  }
}

void
decompose_product (const symbolic& s, symbolic& sc, symbolic_commutative& prod)
{
  if (s->type () == SYMBOLIC_PRODUCT) {
    sc  = SYMB_ONE;
    prod= ((symbolic_product_rep*) access (s)) -> prod;
    return;
  }

  if (s->type () == SYMBOLIC_SUM) {
    symbolic_commutative sum= as_symbolic_sum (s);
    if (N(sum) == 1) {
      symbolic mon= *entries (sum);
      sc= sum [mon];
      symbolic sc2;
      decompose_product (mon, sc2, prod);
      assert (sc2 == SYMB_ONE);
      return;
    }
  }

  if (is_scalar (s)) {
    sc  = s;
    prod= symbolic_commutative (SYMB_ZERO);
  }

  else {
    sc  = SYMB_ONE;
    prod= symbolic_commutative (SYMB_ZERO);
    prod (s)= SYMB_ONE;
  }
}

/******************************************************************************
* Operations
******************************************************************************/

symbolic
product_mul (const symbolic& s1, const symbolic& s2) {
  symbolic sc1, sc2;
  symbolic_commutative prod1, prod2;
  decompose_product (s1, sc1, prod1);
  decompose_product (s2, sc2, prod2);
  return scalar_mul (sc1 * sc2, symbolic_product (prod1 + prod2));
}

symbolic
product_div (const symbolic& s1, const symbolic& s2) {
  symbolic sc1, sc2;
  symbolic_commutative prod1, prod2;
  decompose_product (s1, sc1, prod1);
  decompose_product (s2, sc2, prod2);
  return scalar_mul (sc1 / sc2, symbolic_product (prod1 - prod2));
}

symbolic
product_pow (const symbolic& s1, const symbolic& p) {
  if (s1 == SYMB_ZERO) return SYMB_ZERO;
  else if (s1 == SYMB_ONE) return SYMB_ONE;
  else if (p == SYMB_ZERO) return SYMB_ONE;
  else if (p == SYMB_ONE) return s1;
  else {
    symbolic sc;
    symbolic_commutative prod;
    decompose_product (s1, sc, prod);
    if (/*true*/ N(prod) <= 1) {
      if (sc == SYMB_ONE) return symbolic_product (p * prod);
      else {
	symbolic_commutative sc_pow (SYMB_ZERO);
	sc_pow (sc)= p;
	return symbolic_product (sc_pow + p * prod);
      }
    }
    else {
      symbolic_commutative new_prod (SYMB_ZERO);
      new_prod (s1)= p;
      return new symbolic_product_rep (new_prod);
    }
  }
}

} // namespace mml
